/*
 * Copyright (C) 2009 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *    Cody Russell <crussell@canonical.com>
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <string.h>

#include "gtytimeline.h"
#include "xsplash.h"

typedef struct _AnimContext          AnimContext;
typedef struct _XsplashServerPrivate XsplashServerPrivate;

struct _AnimContext
{
  XsplashServer *server;
  GtkWindow     *window;
  GtyTimeline   *timeline;
  gboolean       starting_opaque;
  gboolean       exit_on_finish;
};

struct _XsplashServerPrivate
{
  gchar           *dbusobject;
  GtkWidget       *window;
  DBusGConnection *system_bus;
  DBusGProxy      *bus_proxy;

  gboolean         nautilus_done;
  gboolean         panel_done;
};

enum {
  PROP_0,
  PROP_DBUS_OBJECT
};

#define XSPLASH_DBUS_NAME   "com.ubuntu.BootCurtain"
#define XSPLASH_DBUS_OBJECT "/com/ubuntu/BootCurtain"

static gboolean gdm_session = FALSE;
static GOptionEntry entries[] = {
  { "gdm-session", 'g', 0, G_OPTION_ARG_NONE, &gdm_session, "Run in gdm session", NULL },
  { NULL }
};

static AnimContext *      anim_context_new               (XsplashServer  *server,
                                                          GtyTimeline    *timeline,
                                                          gpointer        id);
static void               anim_context_free              (AnimContext    *context);
static void               timeline_frame_cb              (GtyTimeline    *timeline,
                                                          gdouble         progress,
                                                          gpointer        user_data);
static void               timeline_finished_cb           (GtyTimeline    *timeline,
                                                          gpointer        user_data);
static void               set_property                   (GObject        *obj,
                                                          guint           id,
                                                          const GValue   *value,
                                                          GParamSpec     *pspec);
static void               get_property                   (GObject        *obj,
                                                          guint           id,
                                                          GValue         *value,
                                                          GParamSpec     *pspec);


static void               window_realized                (GtkWidget      *window,
                                                          gpointer        user_data);
static gboolean           key_press_event                (GtkWidget      *widget,
                                                          GdkEventKey    *event,
                                                          gpointer        user_data);

static void               xsplash_server_dispose         (GObject        *object);

gboolean                  xsplash_server_signal_loaded   (XsplashServer  *server,
                                                          gchar          *app,
                                                          GError        **error);

#include "dbus-xsplash-server.h"

G_DEFINE_TYPE (XsplashServer, xsplash_server, G_TYPE_OBJECT);
#define XSPLASH_SERVER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), XSPLASH_TYPE_SERVER, XsplashServerPrivate))

static gboolean fading = FALSE;

static void
xsplash_server_class_init (XsplashServerClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);

  g_type_class_add_private (class, sizeof (XsplashServerPrivate));

  object_class->dispose = xsplash_server_dispose;
  object_class->set_property = set_property;
  object_class->get_property = get_property;

  g_object_class_install_property (object_class, PROP_DBUS_OBJECT,
                                   g_param_spec_string ("dbus-object", "DBus object path",
                                                        "The object that represents this set of menus on DBus",
                                                        XSPLASH_DBUS_OBJECT,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));

  dbus_g_object_type_install_info (XSPLASH_TYPE_SERVER,
                                   &dbus_glib_xsplash_server_object_info);
}

static void
xsplash_server_dispose (GObject *object)
{
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (object);

  g_object_unref (priv->window);
}

static void
set_property (GObject *obj, guint id, const GValue *value, GParamSpec *pspec)
{
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (obj);

  switch (id) {
  case PROP_DBUS_OBJECT:
    g_return_if_fail (priv->dbusobject == NULL);

    priv->dbusobject = g_value_dup_string (value);

    DBusGConnection *connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);

    dbus_g_connection_register_g_object (connection,
                                         priv->dbusobject,
                                         obj);
    break;

  default:
    g_return_if_reached ();
    break;
  }

  return;
}

static void
get_property (GObject *obj, guint id, GValue *value, GParamSpec *pspec)
{
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (obj);

  switch (id) {
  case PROP_DBUS_OBJECT:
    g_value_set_string (value, priv->dbusobject);
    break;

  default:
    g_return_if_reached ();
    break;
  }

  return;
}

static void
xsplash_server_init (XsplashServer *server)
{
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (server);
  GdkPixbuf *pixbuf;
  GdkScreen *screen;
  GtkWidget *image;
  GError *error = NULL;
  guint nameret = 0;

  priv->dbusobject = NULL;
  priv->system_bus = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);
  priv->bus_proxy = dbus_g_proxy_new_for_name (priv->system_bus,
                                               DBUS_SERVICE_DBUS,
                                               DBUS_PATH_DBUS,
                                               DBUS_INTERFACE_DBUS);

  if (!org_freedesktop_DBus_request_name (priv->bus_proxy,
                                          XSPLASH_DBUS_NAME,
                                          0, &nameret, &error))
    {
      g_error ("Unable to call to request name");
      return;
    }

  if (nameret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
    {
      g_error ("Unable to get name");
      return;
    }

  priv->window = g_object_new (gtk_window_get_type (),
                               "name", "xsplash-window",
                               NULL);
  gtk_window_fullscreen (GTK_WINDOW (priv->window));
  gtk_window_set_keep_above (GTK_WINDOW (priv->window), TRUE);

  gtk_widget_set_colormap (priv->window,
                           gdk_screen_get_rgba_colormap (gdk_screen_get_default ()));

  g_signal_connect (priv->window,
                    "realize",
                    G_CALLBACK (window_realized),
                    server);
  g_signal_connect (priv->window,
                    "key-press-event",
                    G_CALLBACK (key_press_event),
                    server);

  screen = gtk_widget_get_screen (priv->window);

  pixbuf = gdk_pixbuf_new_from_file_at_scale ("/usr/share/backgrounds/warty-final-ubuntu.png",
                                              gdk_screen_get_width (screen),
                                              gdk_screen_get_height (screen),
                                              FALSE,
                                              NULL);

  image = gtk_image_new_from_pixbuf (pixbuf);
  gtk_container_add (GTK_CONTAINER (priv->window), image);

  gtk_widget_show_all (priv->window);
}

static void
begin_fade (XsplashServer *server, gboolean exit_on_finish)
{
  XsplashServerPrivate *priv;
  GtyTimeline *timeline;
  AnimContext *context;

  if (!fading)
    {
      priv = XSPLASH_SERVER_GET_PRIVATE (server);
      timeline = gty_timeline_new (800);
      context = anim_context_new (server,
                                  timeline,
                                  "fade");
      context->exit_on_finish = exit_on_finish;

      g_signal_connect (timeline,
                        "frame",
                        G_CALLBACK (timeline_frame_cb),
                        context);
      g_signal_connect (timeline,
                        "finished",
                        G_CALLBACK (timeline_finished_cb),
                        context);
      gty_timeline_start (timeline);

      fading = TRUE;
    }
}

static gboolean
key_press_event (GtkWidget   *widget,
                 GdkEventKey *event,
                 gpointer     user_data)
{
  XsplashServer *server = (XsplashServer *)user_data;

  switch (event->keyval)
    {
    case GDK_Escape:
      begin_fade (server, TRUE);

      /* else queue another fade? */
      break;

    default:
      break;
    }

  return FALSE;
}

static AnimContext *
anim_context_new (XsplashServer *server,
                  GtyTimeline   *timeline,
                  gpointer       id)
{
  AnimContext *context = g_slice_new (AnimContext);
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (server);

  context->server          = server;
  context->window          = GTK_WINDOW (priv->window);
  context->timeline        = timeline;
  context->starting_opaque = (gtk_window_get_opacity (context->window) == 1.0);

  return context;
}

static void
anim_context_free (AnimContext *context)
{
  g_object_unref (context->timeline);

  g_slice_free (AnimContext, context);
}

static void
timeline_frame_cb (GtyTimeline *timeline,
                   gdouble      progress,
                   gpointer     user_data)
{
  AnimContext *context = user_data;
  gdouble opacity;

  if (context->starting_opaque)
    {
      opacity = 1.0 - progress;
    }
  else
    {
      opacity = progress;
    }

  gtk_window_set_opacity (context->window, 1.0 - progress);
}

static void
timeline_finished_cb (GtyTimeline *timeline,
                      gpointer     user_data)
{
  AnimContext *context = user_data;

  anim_context_free (context);

  if (context->exit_on_finish)
    {
      gtk_main_quit ();
    }
}

static void
window_realized (GtkWidget *window,
                 gpointer   user_data)
{
  XsplashServer *server = (XsplashServer *)user_data;
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (server);

  GdkCursor *cursor = gdk_cursor_new (GDK_BLANK_CURSOR);
  gdk_window_set_cursor (window->window,
                         cursor);
  gdk_cursor_unref (cursor);

  gdk_window_set_type_hint (priv->window->window, GDK_WINDOW_TYPE_HINT_MENU);
}

gboolean
temporary_hack_for_initial_fade (gpointer user_data)
{
  XsplashServer *server = (XsplashServer *)user_data;

  begin_fade (server, TRUE);

  return FALSE;
}

int
main (int argc, char *argv[])
{
  XsplashServer *server;
  GError *error = NULL;
  GOptionContext *context;

  context = g_option_context_new ("xsplash");
  g_option_context_add_main_entries (context,
                                     entries,
                                     "xsplash");
  g_option_context_add_group (context, gtk_get_option_group (TRUE));
  if (!g_option_context_parse (context, &argc, &argv, &error))
    {
      exit (1);
    }

  gtk_init (&argc, &argv);

  server = g_object_new (xsplash_server_get_type (),
                         "dbus-object", XSPLASH_DBUS_OBJECT,
                         NULL);

  if (gdm_session)
    {
      g_timeout_add (10000,
                     temporary_hack_for_initial_fade,
                     server);
    }

  gtk_main ();

  return 0;
}

gboolean
xsplash_server_signal_loaded (XsplashServer  *server,
                              gchar          *app,
                              GError        **error)
{
  XsplashServerPrivate *priv = XSPLASH_SERVER_GET_PRIVATE (server);

  if (gdm_session)
    {
      begin_fade (server, TRUE);
    }
  else
    {
      if (strcmp (app, "nautilus") == 0)
        {
          priv->nautilus_done = TRUE;
        }
      else if (strcmp (app, "gnome-panel") == 0)
        {
          priv->panel_done = TRUE;
        }

      if (priv->nautilus_done && priv->panel_done)
        {
          begin_fade (server, TRUE);
        }
    }

  return TRUE;
}
